;    nuni.asm   A boot loader for linux which does not use the bios.
;               For ext2 file system and IDE drives.
;    Copyright (C) 1999-2000  Neil Koozer
;
;    This program is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;
;    Neil Koozer, 907 SE Sharon Ave, Roseburg, OR 97470-3824
;    nkoozer@rosenet.net
;-------------------------------------------------------------

section .text

ORG 0x7c00

 jmp short Entry	;overlay into unused descriptor 0

gdt	DW 0xf		;15 means gdt is 16 bytes (2 descriptors)
	DD 0x7c00	;absolute address of gdt

			;descriptor 1 starts here
        DW 0FFFFh	;segment limit (bits 0-15)
        DW 0		;base addr (bits 0-15)
        DB 0		;base addr (bits 16-23)
        DB 10010010b	;b7:  1 = segment present
			;b65: privilege level (00)
			;b4:  1 = application type descriptor
			;b0-3:  2 = read/write
        DB 11001111b	;b7:  granularity
			;b6:  default operation size
			;b5:  0 (reserved)
			;b4:  0 (unused)
			;b0-3: segment limit (bits 16-19)
        DB 0		;base addr (bits 24-31)


%include "nuni.dat"

Count	db block_size	;sectors per block
lba	dd 0
	db 0x20		;read command

Entry			;CS:IP = 0:0x7c00  (I hope)
 cli			;it works without cli
 lgdt [gdt]
 mov eax,cr0
 inc al
 mov cr0,eax		;set protected mode
 mov dx,8
 mov es,dx		;make ds & es flat
 mov ds,dx		;need both ds & es for 'movsb'
 dec al
 mov cr0,eax		;set real mode
 mov cx,0x9000
 mov gs,cx		;setup sectors go here
 mov ss,cx		;get stack out of the way
 mov sp,cx		;two birds with one stone
 xor ecx,ecx		;also clear hi bits for later use
 mov ds,cx		;reload so they will work with push/pop
 mov es,cx
 cld			;only need one for whole program

 mov dx,0x3f2		;turn off floppy motor
 sub al,al
 out dx,al

 call kb_wait		;enable a20
 mov al,0xd1
 out 0x64,al
 call kb_wait
 mov al,0xdf
 out 0x60,al
.a2
 inc ax			;wait until a20 works
 mov [0x7c00],ax
 cmp ax,[dword 0x107c00]
 je .a2

 mov DX,iobase2
 mov AL,2		;use polling for IDE device
 out DX,AL

 mov DX,iobase
 add DX,byte 7
.b1 in AL,DX		;wait till controller ready
 and AL,0x88
 jnz .b1

; mov AL,drive & 0xFF	;select drive
; dec DX
; out DX,AL
; inc DX

; mov CX,300		;wait >= 400 ns
;.b2 loop .b2

;.b3 in AL,DX		;wait till controller ready
; and AL,0x88
; jnz .b3

 mov eax,iblock
 lea edi,[Inode]
 push DI
 call ReadBlock		;get block containing inode

 pop DI		
 mov SI,DI
 add SI,ioffset			;byte addr of inode within block
 mov cl,128			;hi bits are 0 already
 rep movsb

%if double_blks
 mov DI,Single+4		;this location fair game for now
 mov eax,[DI]			;get double indirect pointers first
 call ReadBlock
 mov cl,double_blks		;how many blks of double indirect pointers
 mov DI,Double
 mov bx,Single+4
 call ReadLoop
%endif

 mov DI,Single			;get single indirect pointers
 mov eax,[Single]
 call ReadBlock			;now have contiguous array of pointers

 mov edi,0x90000		;load setup blocks here
 mov bx,Direct
 mov cl,setup_blocks
 call ReadLoop

 mov edi,load_address		;where to put kernel
%if move_count
 mov cx,move_count		;in case the last setup block holds
 mov esi,move_start		;a few sectors of kernel code
 rep a32 movsd
%endif

 mov cx,blocks_left		;obstacles cleared, the rest is linear
 call ReadLoop

;done loading the kernel, now load file-based command line
 mov eax,iblock2
 lea edi,[Inode]
 push DI
 call ReadBlock			;get block containing inode
 pop DI		
 mov eax,[DI+ioffset2+40]	;block # of nuni.kcl (command line file)
 mov edi,0x90000+cmd_buff	;where to put it
 call ReadBlock

 mov DX,iobase2
 mov AL,0		;restore int mode for IDE
 out DX,AL

;now allow an interactive command line
 mov BX,cmd_line
 mov AH,2		;check for shift, ctrl, or alt
 int 0x16
 and AL,0xF
 jz short .out		;skip if none pressed
 call PrintThis
 db 10,13,'Cmd: ',0
.more
 mov AH,0
 int 0x16		;get char
 mov [gs:BX],AL
 cmp AL,0x0D
 je short .out
 cmp AL,8
 jne short .f1
 cmp BX,cmd_line
 je short .more
 call PrintThis
 db 8,' ',8,0
 dec BX
 jmp short .more
.f1
 cmp AL,0x20
 jb short .more
 cmp AL,0x7e
 ja short .more
 call PrintChar
 inc BX
 jmp short .more
.out
 call PrintThis
 db 10,13,0
 sub AL,AL
 mov [gs:BX],AL

 mov [gs:0x210],byte 0x50	;type of loader
 mov [gs:0x20],dword 0xa33f + (cmd_buff * 0x10000)
 jmp 0x9020:0

PrintChar
 push BX
 sub BH,BH		;video page 0
 mov AH,0xE
 int 0x10
 pop BX
 ret

PrintThis
 pop SI			;address of string
.more
 mov AL,[CS:SI]
 inc SI
 or AL,AL		;end of asciiz string?
 je short .out
 call PrintChar
 jmp short .more
.out
 jmp SI

ReadLoop
 mov eax,[bx]
 add bx,byte 4
 call ReadBlock
 loop ReadLoop
 ret

ReadBlock
 push CX
 mov CL,blk_shift
 shl eax,CL		;block # to sector #
 add eax,Start		;sector address of partition
 or eax,drive		;append drive# and LBA flag
 mov [lba],eax

 mov SI,Count		;start of six-byte IDE command slate
 mov cx,6		;one command to read a multi-sector ext2 block
 mov DX,iobase
 inc DX
@@b3			;load 6 bytes into controller
 inc DX
 outsb
 loop @@b3
 mov CL,block_size
@@b4
 call data		;each call moves one sector of data
 loop @@b4
 pop CX
print
 mov AL,'.'		;print progress indicator
 jmp short PrintChar

data
 push CX
 mov CX,300		;wait >= 400 ns
.b2 loop .b2
@@b2			;poll for a drq
 in AL,DX
 and AL,88h
 jz error
 cmp AL,8
 jnz @@b2
 push DX
 mov DX,iobase		;get data
 mov cx,0x100		;1 sector = 0x100 words
 rep a32 insw
 pop DX
 pop CX
 ret

error
 mov AL,'E'
 call PrintChar
 jmp short $

kb_wait
 in al,0x64
 and al,2
 jnz kb_wait
 ret

TIMES 510-($-$$) DB 0
 dw 0xaa55
End

section	.bss
Inode	resb 40
Direct	resb 48		;overlay to make contiguous array
Single	resb block_size * 512
Double

